import logging
from collections import namedtuple
from datetime import datetime as dt
from pathlib import Path

##setting up logging
from toargridding.defaultLogging import toargridding_defaultLogging
from toargridding.gridding import get_gridded_toar_data
from toargridding.grids import RegularGrid
from toargridding.metadata import TimeSample
from toargridding.toar_rest_client import AnalysisServiceDownload

logger = toargridding_defaultLogging()
logger.addShellLogger(logging.DEBUG)
logger.logExceptions()
# logger.addRotatingLogFile_scriptName(__file__)
# logger.addSysLogger(logging.DEBUG)

# raise RuntimeError("For testing purposes")

# creation of request.

Config = namedtuple("Config", ["grid", "time", "variables", "stats", "moreOptions"])

varName = "country"
grid = RegularGrid(
    lat_resolution=1.9,
    lon_resolution=2.5,
)

configs = {}
country = "AL"
valid_data = Config(
    grid,
    TimeSample(start=dt(2000, 1, 1), end=dt(2018, 12, 31), sampling="monthly"),  # possibly adopt range:-)
    ["mole_fraction_of_ozone_in_air"],  # variable name
    ["dma8epa"],
    {varName: country},
)

configs[f"test_ta{country}"] = valid_data

# CAVE: this cell runs about 45minutes per requested year. therefore we increase the waiting duration to 1h per request.
# the processing is done on the server of the TOAR database.
# a restart of the cell continues the request to the REST API if the requested data are ready for download
# The download can also take a few minutes

stats_endpoint = "https://toar-data.fz-juelich.de/api/v2/analysis/statistics/"
cache_basepath = Path("cache")
result_basepath = Path("results")
cache_basepath.mkdir(exist_ok=True)
result_basepath.mkdir(exist_ok=True)
analysis_service = AnalysisServiceDownload(
    stats_endpoint=stats_endpoint, cache_dir=cache_basepath, sample_dir=result_basepath, use_downloaded=True
)

# maybe adopt the interval for requesting the results and the total duration, before the client pauses the requests.
# as the requests take about 45min, it is more suitable to wait 60min before timing out the requests than the original 30min.
analysis_service.connection.set_request_times(interval_min=5, max_wait_minutes=60)

createdFiles = []

for config in configs.values():
    try:
        datasets, metadatas = get_gridded_toar_data(
            analysis_service=analysis_service,
            grid=config.grid,
            time=config.time,
            variables=config.variables,
            stats=config.stats,
            **config.moreOptions,
        )
        for _dset in datasets:
            pass
    except KeyError:
        continue
